@@ -12,6 +12,14 @@
 #include <sys/ipc.h>
 #include <sys/sem.h>
 
+typedef struct _GdkIOClosure GdkIOClosure;
+struct _GdkIOClosure
+{
+  GdkInputFunction function;
+  GdkInputCondition condition;
+  GdkDestroyNotify notify;
+  gpointer data;
+};
 
 static void controlSocketCallback(gpointer clientData, gint fd, GdkInputCondition condition);
 static void vncSocketCallback(gpointer clientData, gint fd, GdkInputCondition condition);
@@ -21,10 +29,11 @@ void CheckLockClient(void);
 
 int getClientFromCtldsp (int fd);
 int getClientFromVncdsp (int fd);
-void SendCastMsg (char flag, int clientNum);
+void SendCastMsg (char flag, int clientNum, char fullscreen);
 int ConnectMyself ();
 Bool AuthDB (char *client_ID);
 
+extern int multi_ttl;
 extern GtkWidget *main_window;
 GtkWidget *dialog_IDinput;
 
@@ -197,13 +206,15 @@ AdvInit ()
   }
   server_CB.casting_port = htons(SERVER_PORT);
 
-  sem_init ();
+  //sem_init ();
 }
 
 int
 ConnectMyself ()
 {
-  char tmp[64] = "multivnc_client -d 1 -u ";
+  char tmp[64] = "multivnc_client -d 1 --nonotice -u ";
+  strcat (tmp, server_CB.server_ID);
+  strcat (tmp, " -s ");
   strcat (tmp, server_CB.server_ID);
   strcat (tmp, " &");
   printf("tmp=%s\n",tmp);
@@ -227,8 +238,8 @@ ConnectCTL (struct sockaddr_in client_addr, int ctldsp, int cnum)
   printf("clientip=%s\n",inet_ntoa(client_addr.sin_addr));
   printf("clientport=%d\n",ntohs(client_addr.sin_port));
 
-  server_CB.client_grp[cnum]->g_ctldsp = gdk_input_add (server_CB.client_grp[cnum]->ctldsp, 
-      GDK_INPUT_READ, (GdkInputFunction)controlSocketCallback, NULL);
+  server_CB.client_grp[cnum]->g_ctldsp = gdk_input_add_priority (server_CB.client_grp[cnum]->ctldsp, 
+      GDK_INPUT_READ, (GdkInputFunction)controlSocketCallback, NULL, INPUT_PRIORITY);
 
   printf("ConnectCTL=%d\n",(int)server_CB.client_grp[cnum]);
   printf("ctl=%d, g_ctl=%d\nvnc=%d, g_vnc=%d\n",(int)server_CB.client_grp[cnum]->ctldsp,
@@ -297,14 +308,12 @@ ConnectVNC (struct sockaddr_in client_addr, int vncdsp, int cnum)
 
   clientNum = draw_info[page].client_num[da_num] = cnum;
   CreateImageInfo(FALSE);
-  if ((gdkpixmap[page][da_num]) && (page == current_page) && \
-      (da_num < draw_info[page].row * draw_info[page].column))
+  if ((gdkpixmap[page][da_num]) && (page == current_page) && (da_num < draw_info[page].row * draw_info[page].column) \
+      || (GETSEND(cnum)))
     SendAllFramebufferUpdateRequest();
-//    SendFramebufferUpdateRequest(0, 0, server_CB.client_grp[cnum]->serverInitMsg.framebufferWidth, \
-			        server_CB.client_grp[cnum]->serverInitMsg.framebufferHeight, False);
 
-  server_CB.client_grp[cnum]->g_vncdsp = gdk_input_add (server_CB.client_grp[cnum]->vncdsp, GDK_INPUT_READ, \
-      							(GdkInputFunction)vncSocketCallback, NULL);
+  server_CB.client_grp[cnum]->g_vncdsp = gdk_input_add_priority (server_CB.client_grp[cnum]->vncdsp, GDK_INPUT_READ, \
+							  (GdkInputFunction)vncSocketCallback, NULL, INPUT_PRIORITY);
 
   add_to_dialog_viewset(cnum);
 
@@ -334,27 +343,95 @@ gdk_input_add_all ()
   for (i = 0; i < server_CB.user_num; i++) {
     if (server_CB.client_grp[i] == NULL) continue;
     if (server_CB.client_grp[i]->g_vncdsp != 0)
-    server_CB.client_grp[i]->g_vncdsp = gdk_input_add (server_CB.client_grp[i]->vncdsp,
-                                              GDK_INPUT_READ, (GdkInputFunction)vncSocketCallback, NULL);
+    server_CB.client_grp[i]->g_vncdsp = gdk_input_add_priority (server_CB.client_grp[i]->vncdsp,
+                                              GDK_INPUT_READ, (GdkInputFunction)vncSocketCallback, NULL, INPUT_PRIORITY);
     if (server_CB.client_grp[i]->g_ctldsp != 0)
-    server_CB.client_grp[i]->g_ctldsp = gdk_input_add (server_CB.client_grp[i]->ctldsp,
-		                              GDK_INPUT_READ, (GdkInputFunction)controlSocketCallback, NULL);
+    server_CB.client_grp[i]->g_ctldsp = gdk_input_add_priority (server_CB.client_grp[i]->ctldsp,
+		                              GDK_INPUT_READ, (GdkInputFunction)controlSocketCallback, NULL, INPUT_PRIORITY);
       
   }
 }
 
+static void
+gdk_io_destroy (gpointer data)
+{
+  GdkIOClosure *closure = data;
+
+  if (closure->notify)
+    closure->notify(closure->data);
+
+  g_free(closure);
+}
+
+#define READ_CONDITION (G_IO_IN | G_IO_HUP | G_IO_ERR)
+#define WRITE_CONDITION (G_IO_OUT | G_IO_ERR)
+#define EXCEPTION_CONDITION (G_IO_PRI)
+
+static gboolean
+gdk_io_invoke (GIOChannel   *source,
+	       GIOCondition condition,
+	       gpointer     data)
+{
+  GdkIOClosure *closure = data;
+  GdkInputCondition gdk_cond = 0;
+
+  if (condition & READ_CONDITION)
+    gdk_cond |= GDK_INPUT_READ;
+  if (condition & WRITE_CONDITION)
+    gdk_cond |= GDK_INPUT_WRITE;
+  if (condition & EXCEPTION_CONDITION)
+    gdk_cond |= GDK_INPUT_EXCEPTION;
+
+  if (closure->condition & gdk_cond)
+    closure->function(closure->data, g_io_channel_unix_get_fd(source), gdk_cond);
+
+  return TRUE;
+}
+
+gint
+gdk_input_add_priority (gint              source,
+			GdkInputCondition condition,
+			GdkInputFunction  function,
+			gpointer          data,
+			gint              priority)
+{
+  guint result;
+  GdkIOClosure *closure = g_new(GdkIOClosure, 1);
+  GIOChannel *channel;
+  GIOCondition cond = 0;
+
+  closure->function = function;
+  closure->condition = condition;
+  closure->notify = NULL;
+  closure->data = data;
+
+  if (condition & GDK_INPUT_READ)
+    cond |= READ_CONDITION;
+  if (condition & GDK_INPUT_WRITE)
+    cond |= WRITE_CONDITION;
+  if (condition & GDK_INPUT_EXCEPTION)
+    cond |= EXCEPTION_CONDITION;
+
+  channel = g_io_channel_unix_new(source);
+  result = g_io_add_watch_full(channel, priority, cond, gdk_io_invoke,
+			       closure, gdk_io_destroy);
+  g_io_channel_unref(channel);
+
+  return result;
+}
+
 void
 controlSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition)
 {
   clientNum = getClientFromCtldsp (fd);
-  gdk_input_remove(server_CB.client_grp[clientNum]->g_ctldsp);
   if (clientNum == EMPTY) return;
+  gdk_input_remove(server_CB.client_grp[clientNum]->g_ctldsp);
   int socket = server_CB.client_grp[clientNum]->ctldsp;
   char buf[64];
 
   bzero (buf, 64);
   if (recv (socket, buf, 2, 0) <= 0) {
-    printf("error connection\n");
+    printf("session closed\n");
 
     /* 饤Ⱦõ */
     if ((exDB_flag) && (!chStateDB(clientNum, 0))) {
@@ -365,6 +442,7 @@ controlSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition
       int p,d;
       GdkRectangle gdkrect;
 
+      DestroyImageInfo(clientNum);
       p = server_CB.client_grp[clientNum]->draw_page;
       d = server_CB.client_grp[clientNum]->draw_area;
       draw_info[p].client_num[d] = EMPTY;
@@ -423,7 +501,7 @@ controlSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition
       break;
 
     case rfbClientExist:
-      server_CB.client_grp[clientNum]->state &= 0x3f | (buf[1] & 0xc0);
+      server_CB.client_grp[clientNum]->state &= 0xbf | (buf[1] & 0x40);
 
       struct timeval tmptime;
       gettimeofday(&timenow, 0);
@@ -439,8 +517,8 @@ controlSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition
       printf(" %d\n",buf[0]);
       break;
   } 
-  server_CB.client_grp[clientNum]->g_ctldsp = gdk_input_add (server_CB.client_grp[clientNum]->ctldsp,
-		        GDK_INPUT_READ, (GdkInputFunction)controlSocketCallback, NULL);
+  server_CB.client_grp[clientNum]->g_ctldsp = gdk_input_add_priority (server_CB.client_grp[clientNum]->ctldsp,
+		        GDK_INPUT_READ, (GdkInputFunction)controlSocketCallback, NULL, INPUT_PRIORITY);
   return;
 }
 
@@ -449,6 +527,7 @@ vncSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition)
 {
   sem_lock ();
   clientNum = getClientFromVncdsp (fd);
+  if (clientNum == EMPTY) return;
   gdk_input_remove(server_CB.client_grp[clientNum]->g_vncdsp);
   page = server_CB.client_grp[clientNum]->draw_page;
   da_num = server_CB.client_grp[clientNum]->draw_area;
@@ -458,8 +537,8 @@ vncSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition)
 	(gdkpixmap[page][da_num] == NULL)) {
       char dummybuf[MAXBUF];
       while (read(server_CB.client_grp[clientNum]->vncdsp, dummybuf, MAXBUF) == MAXBUF);
-      server_CB.client_grp[clientNum]->g_vncdsp = gdk_input_add (server_CB.client_grp[clientNum]->vncdsp, \
-						   GDK_INPUT_READ, (GdkInputFunction)vncSocketCallback, NULL);
+      server_CB.client_grp[clientNum]->g_vncdsp = gdk_input_add_priority (server_CB.client_grp[clientNum]->vncdsp, \
+						   GDK_INPUT_READ, (GdkInputFunction)vncSocketCallback, NULL, INPUT_PRIORITY);
       return;
     }
   }
@@ -473,8 +552,8 @@ vncSocketCallback (gpointer clientData, gint fd, GdkInputCondition condition)
   }
   sem_unlock ();
 
-  server_CB.client_grp[clientNum]->g_vncdsp = gdk_input_add (server_CB.client_grp[clientNum]->vncdsp, \
-						GDK_INPUT_READ, (GdkInputFunction)vncSocketCallback, NULL);
+  server_CB.client_grp[clientNum]->g_vncdsp = gdk_input_add_priority (server_CB.client_grp[clientNum]->vncdsp, \
+						GDK_INPUT_READ, (GdkInputFunction)vncSocketCallback, NULL, INPUT_PRIORITY);
   return;
 }
 
@@ -618,7 +697,7 @@ CreateMultiSockToSend (unsigned int ip, int port)
   int sockadv;
   struct sockaddr_in adv;
   struct ip_mreq stMreq;
-  u_long TTLadv = MULTITTL;
+  u_long TTLadv = multi_ttl;
   int optval = 1;
 
   /* Create advertise socket */
@@ -671,7 +750,7 @@ CutVNC (int clientNum)
 }
 
 void 
-SendCastMsg (char flag, int clientNum)
+SendCastMsg (char flag, int clientNum, char fullscreen)
 {
   CastAdv castadv;
   int len;
@@ -682,6 +761,8 @@ SendCastMsg (char flag, int clientNum)
   castadv.flag = flag;
   castadv.casting_ip = server_CB.casting_ip[server_CB.client_grp[clientNum]->recv_grp];
   castadv.casting_port = server_CB.casting_port;
+  castadv.fullscreen = fullscreen;
+  memcpy(&castadv.pixelFormat, &myFormat, sizeof(rfbPixelFormat));
 
   for (i = 0; i < MAXCLIENT; i++) {
     if (server_CB.client_grp[i]->cast_grp == server_CB.client_grp[clientNum]->recv_grp) {
@@ -694,6 +775,19 @@ SendCastMsg (char flag, int clientNum)
   len = write (server_CB.client_grp[clientNum]->ctldsp, &castadv, sizeof(CastAdv));
 }
 
+void 
+SendRecMsg (char flag, int clientNum, int rectime)
+{
+  RecAdv recadv;
+
+  bzero(&recadv, sizeof(RecAdv));
+  recadv.type = rfbRecAdv;
+  recadv.flag = flag;
+  recadv.rectime = rectime;
+
+  write (server_CB.client_grp[clientNum]->ctldsp, &recadv, sizeof(RecAdv));
+}
+
 void 
 SendLockMsg (char flag, int clientNum)
 {
@@ -769,6 +863,7 @@ sem_init ()
 void
 sem_lock ()
 {
+  return;
   /* ޥեμ */
   errno = 0;
   if (semop (semaphore, &semb, 1) == -1) {
@@ -786,6 +881,7 @@ sem_lock ()
 void
 sem_unlock ()
 {
+  return;
   /* ޥեβ */
   semb.sem_op = 1;
   if (semop (semaphore, &semb, 1) == -1) {
